package ctbrec.recorder.download.hls;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.file.Files;
import java.time.Duration;
import java.time.Instant;
import java.util.concurrent.ExecutionException;
import java.util.regex.Pattern;

import javax.xml.bind.JAXBException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.iheartradio.m3u8.ParseException;
import com.iheartradio.m3u8.PlaylistException;

import ctbrec.Config;
import ctbrec.Model;
import ctbrec.Recording;
import ctbrec.io.HttpClient;
import ctbrec.io.StreamRedirectThread;
import ctbrec.recorder.download.ProcessExitedUncleanException;

public class FFmpegDownload extends AbstractHlsDownload {
    private static final transient Logger LOG = LoggerFactory.getLogger(FFmpegDownload.class);

    private transient Config config;
    private transient Process ffmpeg;
    private File targetFile;

    public FFmpegDownload(HttpClient client) {
        super(client);
    }

    @Override
    public void init(Config config, Model model, Instant startTime) {
        this.config = config;
        this.model = model;
        this.startTime = startTime;
        targetFile = config.getFileForRecording(model, "mp4", startTime);
    }

    @Override
    public void start() throws IOException {
        try {
            Files.createDirectories(targetFile.getParentFile().toPath());
            String chunkPlaylist = getSegmentPlaylistUrl(model);

            // @formatter:off
            ffmpeg = Runtime.getRuntime().exec(new String[] {
                    "/usr/bin/ffmpeg",
                    "-y", // overwrite output files without asking
                    "-headers", "User-Agent: " + config.getSettings().httpUserAgent,
                    "-i", chunkPlaylist,
                    "-c", "copy",
                    "-f", "mp4",
                    targetFile.getCanonicalPath()
            });
            // @formatter:on
            int exitCode = 1;
            File ffmpegLog = File.createTempFile(targetFile.getName(), ".log");
            try (FileOutputStream mergeLogStream = new FileOutputStream(ffmpegLog)) {
                Thread stdout = new Thread(new StreamRedirectThread(ffmpeg.getInputStream(), mergeLogStream));
                Thread stderr = new Thread(new StreamRedirectThread(ffmpeg.getErrorStream(), mergeLogStream));
                stdout.start();
                stderr.start();
                exitCode = ffmpeg.waitFor();
                stdout.join();
                stderr.join();
                mergeLogStream.flush();
            }
            if (exitCode == 0) {
                if (ffmpegLog.exists()) {
                    Files.delete(ffmpegLog.toPath());
                }
            } else {
                LOG.info("FFmpeg exit code was {}. Logfile: {}", exitCode, ffmpegLog.getAbsolutePath());
                throw new ProcessExitedUncleanException("FFmpeg exit code was " + exitCode);
            }
        } catch (InterruptedException e) {
            LOG.error("Thread interrupted while waiting for FFmpeg to terminate");
            Thread.currentThread().interrupt();
        } catch (ExecutionException | ParseException | PlaylistException | JAXBException e) {
            LOG.error("Couldn't start FFmpeg process for stream download", e);
        }
    }

    @Override
    public void stop() {
        if (ffmpeg != null && ffmpeg.isAlive()) {
            ffmpeg.destroy();
        }
    }

    @Override
    public Model getModel() {
        return model;
    }

    @Override
    public Duration getLength() {
        return Duration.between(startTime, Instant.now());
    }

    @Override
    public void postprocess(Recording recording) {
        Thread.currentThread().setName("PP " + model.getName());
        try {
            runPostProcessingScript(recording);
        } catch (Exception e) {
            throw new PostProcessingException(e);
        }
    }

    @Override
    public File getTarget() {
        return targetFile;
    }

    @Override
    public String getPath(Model model) {
        String absolutePath = getTarget().getAbsolutePath();
        String recordingsDir = Config.getInstance().getSettings().recordingsDir;
        String relativePath = absolutePath.replaceFirst(Pattern.quote(recordingsDir), "");
        return relativePath;
    }

    @Override
    void internalStop() {
        stop();
    }

}
